/* Jatha -- bringing to Java some of the power of lisp.
 * Copyright (C) 2001  Kimberley Burchett   kimbly@pobox.com
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jatha;

import java.io.*;
import java.util.*;

/**
 * This is the class that does the actual macro expansion.  It can get its 
 * input from a file or from a LineNumberReader.
 * 
 * <p>Note: when an Expander is created, it has no registered MacroFactories.
 * If you want it to use a DefaultMacroFactory, you'll have to register it
 * yourself.
 */
public class Expander {
	private Hashtable m_messages = new Hashtable();
	private File m_curFile;
	private int m_curLine = -1;
	// these don't get cleared out when a new file is started.
	private Vector m_globalMacroFactories = new Vector();
	// these get cleared out at the beginning of a new file.
	private Vector m_fileMacroFactories = new Vector();

	public void processFile(File inFile, File outFile, 
		boolean force, boolean quiet)
	{
		String inFilename = inFile.getName();
		String outFilename = outFile.getName();
		m_curFile = inFile;

		if (outFile.exists()) {
			// if the output file is up to date, do nothing
			if (!force && (outFile.lastModified() > inFile.lastModified())) {
				if (!quiet) {
					System.out.print(outFilename+" is up to date; ");
					System.out.println("use -all to force regeneration");
				}
				return;
			}
			// otherwise, delete it
			if (!outFile.delete())
				error("Cannot delete file "+outFilename);
		}

		try {
			LineNumberReader in = new LineNumberReader(new FileReader(inFile));
			FileWriter out = new FileWriter(outFile);

			// try-finally block ensures that the files are closed before we 
			// try to handle exceptions.  Some exception handlers try to delete 
			// the output file, which won't work if we still have it open.
			try {
				processFile(in, out);
			} finally {
				out.flush();
				out.close();
				in.close();
				out = null;
				in = null;
			}
		}
		catch (FileNotFoundException e) {
			error("Cannot find file: "+e.getMessage());
		}
		catch (IOException e) {
			error("IO Error: "+e.getMessage());
		}
		catch (MacroError e) {
			/* don't print error message -- it's already been printed */
			if (!outFile.delete()) {
				System.out.println("Unable to delete file "+outFilename+" after error was encountered.");
				System.out.println(outFilename+" may contain errors.");
			}
		}
		m_messages.clear();
	}

	public void processFile(LineNumberReader in, Writer out) 
		throws IOException, MacroError
	{
		// read one character at a time until end of file
		for (int c = in.read(); c != -1; c = in.read()) {
			// if this is a normal character (not '@'), just pass it 
			// straight from input to output.
			if (c != '@') {
				out.write((char)c);
				continue;
			}

			c = in.read();
			// If we find two consecutive @'s, this isn't a macro;
			// just spit one @ to the output.
			if (c == '@') {
				out.write((char)c);
				continue;
			}

			// we found a macro -- read the macro name
			m_curLine = in.getLineNumber()+1;	// LineNumberReader is 0-based
			StringBuffer name = new StringBuffer("");
			while (c != -1) {
				// we haven't gotten to the macro arguments yet have we?
				if (c != '(') {
					if (!Character.isLetter((char)c) && (c != '_') && (c != '.')) {
						// oops -- we got to the end of the macro name
						// and there was no open parenthesis.  This means
						// that it wasn't a macro after all -- it's
						// probably a javadoc tag.  So spit everything
						// we've consumed so far, unmolested.
						out.write('@');
						out.write(name.toString());
						out.write(c);
						// break the loop that reads the macro name.
						// TBD: we have a bug here if the current character
						// is an @ that starts a real macro.  But that's
						// unlikely, so I'm ignoring it.
						break;
					}
					name.append((char)c);
					c = in.read();
					continue;
				}

				// we have gotten to the macro arguments
				int parenDepth = 1;	// # of ')' characters expected
				StringBuffer args = new StringBuffer("");
				for (c = in.read(); c != -1; c = in.read()) {
					// have we found the end of the arguments?
					if ((c == ')') && (--parenDepth == 0))
						break;	// don't include last ')' in args
					else if (c == '(')
						parenDepth++;
					args.append((char)c);
				}
				// if we're not at the end of the file, we must
				// have made it to the end of the arguments.
				if (c != -1)
					expandMacro(name.toString(), args.toString(), out);
				// break the loop that reads the macro name.
				break;
			}
			// we've either finished the macro or hit the end of file
			if (c == -1)
				error("Unexpected end of file while expanding macro "+name);
			m_curLine = -1;
		}
	}

	public void expandMacro(String macroName, String args, Writer out) {
		for (int i=0; i < m_globalMacroFactories.size(); ++i) {
			MacroFactory factory = (MacroFactory)m_globalMacroFactories.elementAt(i);
			Macro macro = factory.createMacro(macroName, this);
			if (macro != null) {
				try {
					macro.expand(args, out, this);
				} catch (IOException e) {
					error("I/O error during expansion of \""+macroName+"\" macro: "+
						e.getMessage());
				}
				return;
			}
		}
		error("Unknown macro \""+macroName+"\"");
	}

	/**
	 * Registers a new MacroFactory which will be used to lookup Macros.
	 * 
	 * @param justThisFile if true, the macro factory will only be in effect
	 *        for the current file.
	 */
	public void registerMacroFactory(MacroFactory factory, 
		boolean justThisFile)
	{
		if (unregisterMacroFactory(factory))
			trace("macro factory "+factory.getClass().getName()+" registered twice");
		if (justThisFile)
			m_fileMacroFactories.addElement(factory);
		else
			m_globalMacroFactories.addElement(factory);
	}

	/**
	 * Unregister a macro factory.  Returns false if the factory was not
	 * registered.
	 */
	public boolean unregisterMacroFactory(MacroFactory factory) {
		return m_globalMacroFactories.removeElement(factory)
		    || m_fileMacroFactories.removeElement(factory);
	}

	/**
	 * Returns the name of the file currently being processed.  If the version 
	 * of processFile() that takes a filename was not called, returns null 
	 * (ie if you called the version of processFile() that just takes a 
	 * LineNumberReader).
	 */
	public File getFile() {
		return m_curFile;
	}

	/**
	 * Returns the current line number in the current file.  This is determined
	 * from the LineNumberReader passed to processFile(), but is adjusted so 
	 * the first line is line 1, instead of line 0.
	 */
	public int getLine() {
		return m_curLine;
	}

	/**
	 * Allows a macro to leave a message that can be retrieved by later macros.
	 */
	public void leaveMessage(String description, Object message) {
		m_messages.put(description, message);
	}

	/**
	 * Allows a macro to retrieve a message left by a previous macro.
	 */
	public Object getMessage(String description) {
		return m_messages.get(description);
	}

	/**
	 * Prints an error message and throws a MacroError.
	 */
	public void error(String msg) {
		trace(msg);
		throw new MacroError(msg);
	}

	/**
	 * Prints an error message, but doesn't throw a MacroError like error() does.
	 */
	public void trace(String msg) {
		if (getFile() == null) {
			/* don't print file or line */
		}
		else if (getLine() == -1)
			System.out.print(getFile()+": ");
		else
			System.out.print(getFile()+"("+getLine()+"): ");
		System.out.println(msg);
		System.out.flush();
	}
}
