/* Jatha -- bringing to Java some of the power of lisp.
 * Copyright (C) 2001  Kimberley Burchett   kimbly@pobox.com
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jatha;

import java.util.Vector;

/**
 * Helper functions for macros.
 */
public class MacroUtil {
	/**
	 * Capitalize the first character in a String.
	 */
	public static String capitalize(String str) {
		return Character.toUpperCase(str.charAt(0)) + str.substring(1);
	}

	/**
	 * Turn a singular noun into a plural noun.  This works for most English 
	 * words, but not all of them.  The algorithm is simple: if the word ends 
	 * in 's', "es" is stuck on the end.  Otherwise, "s" is stuck on.  So 
	 * "dog" becomes "dogs", and "pass" becomes "passes".
	 */
	public static String pluralize(String str) {
		if (str.endsWith("s"))
			return str+"es";
		else
			return str+"s";
	}

	/**
	 * Split a String into an array of Strings, using a given separator.
	 * 
	 * <p>Examples:
	 * <pre>
	 * String str = "a(x ,y)b \"c,d\" ,"
	 * splitString(str, ',', true, true, true, true)  == {"a(x ,y)b \"c,d\"", ""}
	 * splitString(str, ',', false, true, true, true) == {"a(x ,y)b \"c,d\" ", ""}
	 * splitString(str, ',', true, false, true, true) == {"a(x ,y)b \"c,d\""}
	 * splitString(str, ',', true, true, false, true) == {"a(x", "y)b \"c,d\"", ""}
	 * splitString(str, ',', true, true, true, false) == {"a(x ,y)b \"c", "d\"", ""}
	 * </pre>
	 * 
	 * @param trim if true, trims each element in the result.
	 * @param allowEmpty if false, empty strings are not included in result.
	 * @param nestParens if true, separators inside parentheses are ignored.
	 * @param stringMode if true, separators inside quotes are ignored.
	 */
	public static String[] splitString(String str, char separator, 
		boolean trim, boolean allowEmpty, boolean nestParens, boolean stringMode)
	{
		int length = str.length();
		Vector vectResult = new Vector();
		int parenDepth = 0;
		boolean inString = false;
		int startIndex = 0;		// start index of current string segment
		int i;
		for (i=0; i < length; ++i) {
			char c = str.charAt(i);
			if (nestParens && c == '(') parenDepth++;
			else if (nestParens && c == ')') parenDepth--;
			else if (stringMode && c == '\"') inString = !inString;
			else if (c == separator && 
				(parenDepth == 0 || !nestParens) &&
				(!inString || !stringMode))
			{
				String elem = str.substring(startIndex, i);
				startIndex = i+1;
				if (trim)
					elem = elem.trim();
				if (allowEmpty || elem.length() > 0)
					vectResult.addElement(elem);
			}
		}
		// add the last section to the result
		if (startIndex != i) {
			String elem = str.substring(startIndex, i);
			if (trim)
				elem = elem.trim();
			if (allowEmpty || elem.length() > 0)
				vectResult.addElement(elem);
		}
		String[] result = new String[vectResult.size()];
		vectResult.copyInto(result);
		return result;
	}

	/**
	 * Searches through a string to find a matching ')' character.  Returns 
	 * -1 if no matching ')' exists.  For example, if str is "a(b())+c(", 
	 * and index is 1, then the result is 5.  If index is 3, the result is 4.
	 * If index is 8, the result is -1.  If the character at the given 
	 * startIndex is not a '(' character (e.g. 0,2,4-7), the result is -1.
	 */
	public static int findMatchingEndParen(String str, int startIndex) {
		if (str.charAt(startIndex) != '(')
			return -1;
		int length = str.length();
		for (int i=startIndex+1; i < length; ++i) {
			if (str.charAt(i) == ')')
				return i;
			else if (str.charAt(i) == '(') {
				// skip nested parens and continue
				i = findMatchingEndParen(str, i);
				if (i == -1)
					return -1;
			}
		}
		return -1;
	}

	/**
	 * Replaces special characters like newlines, tabs, quotes, and backslashes
	 * into escaped character sequences (e.g. "\n", "\t", "\\"), and turns 
	 * unportable characters (&lt;32 or &gt;127) into unicode escapes (e.g. 
         * "\u0128").
	 */
	public static String escapify(String str) {
		StringBuffer result = new StringBuffer();
		int length = str.length();
		char[] chars = str.toCharArray();
		for (int i=0; i < length; ++i) {
			switch (chars[i]) {
			case '\\':	result.append("\\\\");	break;
			case '\"':	result.append("\\\"");	break;
			case '\'':	result.append("\\\'");	break;
			case '\b':	result.append("\\b");	break;
			case '\f':	result.append("\\f");	break;
			case '\n':	result.append("\\n");	break;
			case '\r':	result.append("\\r");	break;
			case '\t':	result.append("\\t");	break;
			default:
				// turn unportable chars into unicode escapes
				if (chars[i] < 32 || chars[i] > 127) {
					String hex = Integer.toString(chars[i], 16);
					result.append("\\u");
					for (int j=hex.length(); j < 4; ++j)
						result.append('0');
					result.append(hex);
				}
				else
					result.append(chars[i]);
				break;
			}
		}
		return result.toString();
	}

	/**
	 * Returns true if the string is a Java reserved word.  For example,
	 * "this", "if", "static", or "package".
	 */
	public static boolean isJavaReservedWord(String str) {
		return str.equals("abstract")   || str.equals("this")
		    || str.equals("default")    || str.equals("if")
		    || str.equals("private")    || str.equals("throw")
		    || str.equals("boolean")    || str.equals("do")
		    || str.equals("implements") || str.equals("protected")
		    || str.equals("throws")     || str.equals("break")
		    || str.equals("double")     || str.equals("import")
		    || str.equals("public")     || str.equals("transient")
		    || str.equals("byte")       || str.equals("else")
		    || str.equals("instanceof") || str.equals("return")
		    || str.equals("try")        || str.equals("case")
		    || str.equals("extends")    || str.equals("int")
		    || str.equals("short")      || str.equals("void")
		    || str.equals("catch")      || str.equals("final")
		    || str.equals("interface")  || str.equals("static")
		    || str.equals("volatile")   || str.equals("char")
		    || str.equals("finally")    || str.equals("long")
		    || str.equals("super")      || str.equals("while")
		    || str.equals("class")      || str.equals("float")
		    || str.equals("native")     || str.equals("switch")
		    || str.equals("const")      || str.equals("for")
		    || str.equals("new")        || str.equals("synchronized")
		    || str.equals("continue")   || str.equals("goto")
		    || str.equals("package");
	}

	/**
	 * Just like String.charAt(), but it counts from the end of the string.
	 * reverseCharAt("foobar", 0) returns 'r'.
	 */
	public static char reverseCharAt(String str, int index) {
		return str.charAt(str.length() - index - 1);
	}

	/**
	 * Removes characters from both ends of a string.  clip("foobar", 1, 2)
	 * would return "oob".
	 */
	public static String clip(String str, int front, int back) {
		return str.substring(front, str.length()-back);
	}
}
