/* Jatha -- bringing to Java some of the power of lisp.
 * Copyright (C) 2001  Kimberley Burchett   kimbly@pobox.com
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package macros;

import java.io.*;
import jatha.*;

/**
 * <code>PROP(type, name, [initial-value])</code>
 * 
 * <p>The PROP (short for property) macro declares a private member variable, 
 * a get-function, and a set-function.  Example:  <code>@PROP(String, name)</code> 
 * will generate code something like this:
 * 
 * <pre>
 * private String m_name;
 * public String getName() { return m_name; }
 * public void setName(String name) { m_name = name; }
 * </pre>
 * 
 * <p>The "m_" prefix is primarily to ensure that if you choose a name that
 * happens to be a java reserved word (for example, "interface"), the name
 * of the variable won't cause a compile error.
 * 
 * <p>Note: if the type is "boolean", the get function uses a prefix of "is" 
 * instead of "get" (e.g. isHungry() instead of getHungry()).
 *
 * <p>Also note: the name of the property is automatically capitalized for
 * the get/set functions.  So <code>@PROP(String, name)</code> generates a 
 * function called <code>getName()</code>, not <code>getname()</code>.  If
 * you capitalize the name of the property, the only things affected will be
 * the name of the member variable and the name of the parameter to the set
 * function.
 * 
 * <p>Also note: if you write your own macros, you should know that this 
 * macro leaves a message recording the type of any property it declares.
 * For example, <code>getMessage("PROP`typeof`myProperty")</code> will 
 * return the String type of the "myProperty" property.  Also, if you 
 * need to reference the member variable that stores a property's value, 
 * you shouldn't assume that it's simply "m_"+propertyName; instead, use 
 * getNameOfMemberForProperty().
 */
public class PROP extends Macro {
	/** Expands the macro */
	public void expand(String[] args, Writer out, Expander expander) throws IOException {
		if (args.length != 2 && args.length != 3)
			expander.error("@PROP requires either two or three arguments");
		String type = args[0];
		String name = args[1];
		String caps = util.capitalize(name);
		String init = (args.length == 3) ? " = "+args[2] : "";
		String member = getNameOfMemberVariable(name);
		String getMethod = getNameOfGetMethod(type, name);
		String param = getNameOfParameter(type, name);

		out.write("private "+type+" "+member+init+";\n");
		out.write("\tpublic "+type+" "+getMethod+"() { return "+member+"; }\n");
		out.write("\tpublic void set"+caps+"("+type+" "+param+") { "+member+" = "+param+"; }");

		expander.leaveMessage("PROP`typeof`"+name, type);
	}

	/**
	 * Returns an appropriate name for a method parameter, given the name 
	 * and type of a property.  For example, a boolean property "foo" will
	 * return "isFoo".  A property called "class" will return "theClass".
	 * Most parameter names, however, are the same as the property name --
	 * an int property called "bar" will just return "bar".
	 */
	public static String getNameOfParameter(String type, String propertyName) {
		if (type.equals("boolean"))
			return "is"+util.capitalize(propertyName);
		else if (util.isJavaReservedWord(propertyName))
			return "the"+util.capitalize(propertyName);
		else
			return propertyName;
	}

	/**
	 * Returns the name of the get method for the given property name and
	 * type.  If the type is boolean, the method will start with "is",
	 * otherwise, it starts with "get".
	 */
	public static String getNameOfGetMethod(String type, String propertyName) {
		if (type.equals("boolean"))
			return "is"+util.capitalize(propertyName);
		return "get"+util.capitalize(propertyName);
	}

	/**
	 * Returns the name of the member variable used to store the value of 
         * the given property.
	 */
	public static String getNameOfMemberVariable(String propertyName) {
		return "m_"+propertyName;
	}
}
